const gulp = require('gulp');
const del = require('del');
const chalk = require('chalk');
const tap = require('gulp-tap');
const file = require('gulp-file');
const gulpif = require('gulp-if');
const watch = require('gulp-watch');
const babel = require('gulp-babel');
const rename = require('gulp-rename');
const replace = require('gulp-replace');
const foreach = require('gulp-foreach');
const gap = require('gulp-append-prepend');
const runSequence = require('run-sequence');

const fs = require('fs');
const path = require('path');
const axios = require('axios');
const moment = require('moment');
const { argv } = require('yargs');
const autopolyfiller = require('autopolyfiller');

const pretty = require('pretty-hrtime');
const through = require('through2');

const prefix = '[🐟] ';

const IN_DIRECTORY = process.cwd();
const OUT_DIRECTORY = process.cwd();

// console.log(__dirname);
let HOT = false;
let ERROR = false;
let FIRSTRUN = true;
let TOTAL_TIME = 0;
// Set Global arguments
const ARGUMENTS = {};

let includes = {};
const polyfills = [];

function swallowError(err) {
  // eslint-disable-next-line
  console.log(err.message || err);
  ERROR = true;
  return false;
}

function duration(name = 'gulp-duration') {
  let start = process.hrtime();
  const stream = through.obj({
    objectMode: true,
  });

  function resetStart() {
    start = process.hrtime();
  }

  function log(str) {
    if (!ERROR) {
      // eslint-disable-next-line
      console.log(prefix + str);
    }
  }

  stream.start = resetStart;

  return stream.once('end', () => {
    const rawTime = process.hrtime(start);
    const time = pretty(rawTime);
    // eslint-disable-next-line
    TOTAL_TIME += rawTime[0] * 1000 + rawTime[1] / 1000000;
    log(`${name}: ${chalk.magenta(time)}`);
  });
}

const setArgs = () => {
  let fileArgs = {};
  try {
    fileArgs = JSON.parse(fs.readFileSync(path.join(IN_DIRECTORY, '.goldfish.json'), 'utf8'));
  } catch (e) {
    // eslint-disable-next-line
    console.log("");
  }
  ARGUMENTS.EXT = argv.ext || fileArgs.ext || 'es6';
  ARGUMENTS.DEST = argv.out || fileArgs.out || 'dist';
  ARGUMENTS.SOURCE = argv.in || fileArgs.in || '';
  ARGUMENTS.HOST =
    (argv.host && `http://${argv.host}/v2/`) ||
    (fileArgs.host && `http://${fileArgs.host}/v2/`) ||
    'https://cdn.polyfill.io/v2/';
  ARGUMENTS.MIN = argv.min || fileArgs.min || false;
  ARGUMENTS.NODEPS = argv.nodeps || fileArgs.nodeps || true;
  ARGUMENTS.NOLOG = argv.nolog || fileArgs.nolog || true;
};

// Set default arguments
setArgs();

const build = (args) => {
  if (args) {
    setArgs();
  }
  ERROR = false;
  TOTAL_TIME = 0;
  if (FIRSTRUN) {
    // eslint-disable-next-line
    console.log(
      chalk.white(
        `${prefix}Diving in! Hooking`,
        path.join(IN_DIRECTORY, ARGUMENTS.SOURCE, `*.${ARGUMENTS.EXT}...`),
      ));
    FIRSTRUN = false;
  }

  runSequence(
    'include',
    'compile',
    'detect',
    'asm',
    'modify',
    'polyfill',
    'copy',
    'clean',
  );
};

gulp.task('include', () => {
  // Include:
  const includePath = path.join(
    IN_DIRECTORY,
    ARGUMENTS.SOURCE,
    `**/*.${ARGUMENTS.EXT}`,
  );
  const includeRegex = /include\s*\((?:`|'|")(.+)(?:`|'|")\s*\);*/;
  return gulp
    .src([
      includePath,
      `!${IN_DIRECTORY}/*(_temp_|node_modules)/**/*.${ARGUMENTS.EXT}`,
    ])
    .pipe(foreach((stream, f) => {
      const filePath = path.parse(f.path);
      return stream.pipe(replace(new RegExp(includeRegex, 'g'), (match) => {
        const key = `${filePath.dir}/${filePath.name}`;
        if (!includes[key]) {
          includes[key] = [];
        }
        includes[key].push(`<!--#include file="${match.match(includeRegex)[1]}"-->`);
        return '';
      }));
    }))
    .pipe(duration('Generating includes'))
    .pipe(gulp.dest(path.join(__dirname, '_temp_')));
});

gulp.task('compile', () =>
  gulp
    .src(path.join(__dirname, '_temp_', `**/*.${ARGUMENTS.EXT}`))
    .pipe(foreach(stream =>
      stream.pipe(babel({
        minified: ARGUMENTS.MIN,
        comments: !ARGUMENTS.MIN,
        highlightCode: true,
        presets: [
          [
            require.resolve('@babel/preset-env'),
            {
              shippedProposals: true,
              targets: {
                browsers: ['ie 8'],
              },
            },
          ],
          [
            require.resolve('@babel/preset-stage-2'),
            {
              decoratorsLegacy: true,
            },
          ],
        ],
        plugins: [
          [require.resolve('@babel/plugin-proposal-object-rest-spread'), { loose: true }],
          [require.resolve('@babel/plugin-syntax-object-rest-spread'), { loose: true }],
          [require.resolve('@babel/plugin-proposal-optional-chaining'), { loose: true }],
          [require.resolve('@babel/plugin-proposal-class-properties'), { loose: true }],
          [require.resolve('@babel/plugin-transform-classes'), { loose: true }],
        ],
        // eslint-disable-next-line
          }).on("error", function(err) {
        swallowError(err);
        this.emit('end');
      }))))
    // eslint-disable-next-line
    .on("error", function() {
      this.emit('end');
    })
    .pipe(duration('Compiling to ES3'))
    .pipe(gulp.dest(path.join(__dirname, '_temp_'))));

gulp.task('detect', () =>
  gulp
    .src(path.join(__dirname, '_temp_', '**/*.js'))
    .pipe(foreach((stream, f) =>
      stream.pipe(tap(() => {
        autopolyfiller()
          .add(f.contents.toString('utf-8'))
          .polyfills.forEach((poly) => {
            if (polyfills.indexOf(poly) === -1) {
              polyfills.push(poly);
            }
          });
      }))))
    .pipe(duration('Detecting needed polyfills')));

gulp.task('asm', () =>
  gulp
    .src(path.join(__dirname, '_temp_', '**/*.js'))
    .pipe(foreach(stream => stream.pipe(replace(/\/\/\*\s*/, "")))) //eslint-disable-line
    .pipe(gulp.dest(path.join(__dirname, '_temp_'))));

gulp.task('modify', () =>
  gulp
    .src(path.join(__dirname, '_temp_', '**/*.js'))
    .pipe(foreach((stream, f) =>
      stream
        .pipe(gap.prependText(`<!--#include file="${path.join(
          path.relative(
            path.parse(f.path).dir.replace('/_temp_', ''),
            __dirname,
          ),
          '_pf.asp',
        )}"-->
${
  ARGUMENTS.NODEPS
    ? fs
      .readdirSync(path.join(__dirname, '/deps'))
      .map(dep =>
        `<!--#include file="${path.join(
          path.relative(
            path.parse(f.path).dir.replace('/_temp_', ''),
            __dirname,
          ),
          dep,
        )}"-->`)
    : ''
}
${
  includes[
    path.join(
      IN_DIRECTORY,
      ARGUMENTS.SOURCE,
      path.relative(
        path.join(__dirname, '_temp_'),
        `${path.parse(f.path).dir}/${path.parse(f.path).name}`,
      ),
    )
  ]
    ? includes[
      path.join(
        IN_DIRECTORY,
        ARGUMENTS.SOURCE,
        path.relative(
          path.join(__dirname, '_temp_'),
          `${path.parse(f.path).dir}/${path.parse(f.path).name}`,
        ),
      )
    ].join('\n')
    : ''
}
<script language="javascript" runat="server">
//Built ${moment().format('DD-MMM-YY, hh:mm:ss A')}, ${
  Intl.DateTimeFormat().resolvedOptions().timeZone
}\n`))
        .pipe(replace(/\/\/#.+\s+/g, () => ''))
        .pipe(gap.appendText('</script>'))
        .pipe(gulpif(
          ARGUMENTS.NOLOG,
          replace('console.log(', () => 'Response.Write('),
        ))
        .pipe(rename((pF) => {
              pF.extname = ".asp"; //eslint-disable-line
        }))))
    .pipe(duration('Preparing to polyfill'))
    .pipe(gulp.dest(path.join(OUT_DIRECTORY, ARGUMENTS.DEST))));

const dev = () =>
  watch(
    [path.join(IN_DIRECTORY, ARGUMENTS.SOURCE, `**/*.${ARGUMENTS.EXT}`)],
    {
      ignoreInitial: false,
    },
    () => {
      HOT = true;
      build();
    },
  );

gulp.task('polyfill', () => {
  const polyfillTimer = duration('Generating polyfills');
  // Polyfill:
  if (polyfills.length > 0) {
    return new Promise((resolve) => {
      axios
        .get(`${ARGUMENTS.HOST}polyfill.js?features=${polyfills.join(',')}&excludes=Object.defineProperty&ua=ie/8`)
        .then(({ data }) => {
          resolve();
          return file(
            '_pf.asp',
            `<script language="javascript" runat="server">\n// Polyfilling ${polyfills.join(', ')}\n${data}\n</script>`,
            {
              src: true,
            },
          )
            .pipe(polyfillTimer)
            .pipe(gulp.dest(path.join(OUT_DIRECTORY, ARGUMENTS.DEST)));
        });
    });
  }
  return file(
    '_pf.asp',
    '<script language="javascript" runat="server">\n// No polyfills... yet?\n</script>',
    {
      src: true,
    },
  )
    .pipe(polyfillTimer)
    .pipe(gulp.dest(path.join(OUT_DIRECTORY, ARGUMENTS.DEST)));
});

gulp.task('copy', () =>
  gulp
    .src(path.join(__dirname, '/deps/**'))
    .pipe(gulpif(
      ARGUMENTS.NODEPS,
      gulp.dest(path.join(OUT_DIRECTORY, ARGUMENTS.DEST)),
    ))
    .pipe(duration('Exporting to .asp')));

gulp.task('clean', () => {
  // Delete temps:

  if (HOT) {
    includes = {};
    // eslint-disable-next-line
    console.log("-----------------------------------");
    // eslint-disable-next-line
    console.log(
      `${prefix}Watching for changes at`,
      `${path.join(IN_DIRECTORY, ARGUMENTS.SOURCE)}/...`,
    );
  }

  if (ERROR) {
    // eslint-disable-next-line
    console.log(chalk.white.bgRed.bold("\n   ERROR in compile    "));
  } else {
    const strDone = chalk.white.bgGreen.bold(`\n   Done in ${(TOTAL_TIME / 1000).toFixed(2)}s   `);
    // eslint-disable-next-line
    console.log(strDone);
  }
  return del([path.join(__dirname, '_temp_')], { force: true });
});

module.exports = {
  build,
  dev,
};
